(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009, 2010)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2010-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)
(*
high-level commands and declarations

we need a better name
*)

signature TERM_FUN =
sig

  val register_datatypes : string list -> theory -> theory

  val setup : theory -> theory
end

structure Term_Fun : TERM_FUN =
struct

fun register_datatypes tnames =
  Embed.define_embedding (hd tnames)
  #> fold Trs_Data.declare_constrs tnames
  #> fold Isafor_Tactics.add_constr_sims tnames

open Conv
open Isafor_Basics


fun define_trs prove_sim raw_simps thy =
  let
    val ctxt = Proof_Context.init_global thy
    val simps = Attrib.eval_thms ctxt raw_simps
    val conv : conv = (concl_conv ~1 o arg_conv o arg_conv)
      (Simplifier.rewrite (put_simpset HOL_basic_ss ctxt addsimps (Trs_Data.Subst.get ctxt)))
    val simps' = simps |> map (fconv_rule conv)
                 
    fun add_sim_rule thy =
      let
        val sim_rule =
          Isafor_Tactics.derive_sim_lemma_any (Proof_Context.init_global thy) simps'
          |> Thm.close_derivation
      in
        Context.theory_map (Trs_Data.add_sim sim_rule) thy
      end
  in
    thy
    |> Trs_Data.define_trs simps'
    |> (if prove_sim then add_sim_rule else I)
  end

fun define_dps raw_simps thy =
  let
    val ctxt = Proof_Context.init_global thy
    val simps = Attrib.eval_thms ctxt raw_simps
    val conv : conv = (concl_conv ~1 o arg_conv o arg_conv)
      (Simplifier.rewrite (put_simpset HOL_basic_ss ctxt addsimps (Trs_Data.Subst.get ctxt)))
    val simps' = simps |> map (fconv_rule conv)
                 
  in
    thy
    |> Trs_Data.define_dps simps'
  end


fun ceta_import_proof_tac ctxt path i = 
  let 
    val prf = File.read path
       |> Ceta_Parser.read_cpf
       |> Thm.cterm_of ctxt
  in 
    resolve_tac ctxt @{thms Ceta_Checker.ceta_checker} i
    THEN PRIMITIVE (Thm.instantiate' [] [SOME prf])  
  end

fun ceta_import_dp_proof_tac ctxt path i = 
  let 
    val prf = File.read path
       |> Ceta_Parser.read_dp_cpf
       |> Thm.cterm_of ctxt
  in 
    resolve_tac ctxt @{thms Ceta_Checker.ceta_dp_checker} i
    THEN PRIMITIVE (Thm.instantiate' [] [SOME prf])  
  end

val ceta_checker_simp_tac = Code_Simp.static_tac 
  {ctxt = @{context}, simpset = NONE, consts = [@{const_name Ceta_Checker.ceta_checker}]}

fun extract_trs_problem goal_state =
  let
    val subgoal = hd (Thm.prems_of goal_state)
    val (qL,trsL) = case subgoal of 
      Const (@{const_name HOL.Trueprop}, _)
        $ (Const (@{const_name Abstract_Rewriting.SN_on}, _)
        $ (Const (@{const_name Q_Restricted_Rewriting.qrstep}, _)
        $ @{term True}
          $ qL
          $ (Const (@{const_name List.set}, _) $ trsL))
        $ _ (*UNIV*)) => (qL,trsL)
    | _ => raise TERM ("Bad goal", [subgoal])
  in (qL,trsL)
  end

fun extract_dp_problem goal_state = 
  let
    val subgoal = hd (Thm.prems_of goal_state)
    val (qL,pL,rwL) = case subgoal of 
      Const (@{const_name HOL.Trueprop}, _)
        $ (Const (@{const_name QDP_Framework.finite_dpp}, _) $
        (Const (@{const_name Product_Type.Pair}, _) $ @{term True} $
        (Const (@{const_name Product_Type.Pair}, _) $ @{term False} $
        (Const (@{const_name Product_Type.Pair}, _) $ (Const (@{const_name List.set}, _) $ pL) $
        (Const (@{const_name Product_Type.Pair}, _) $ (Const (@{const_name List.set}, _) $ _) $
        (Const (@{const_name Product_Type.Pair}, _) $ qL $
        (Const (@{const_name Product_Type.Pair}, _) $ (Const (@{const_name List.set}, _) $ _) $
        (Const (@{const_name List.set}, _) $ rwL))))))))
        => (qL,pL,rwL)
    | _ => raise TERM ("Bad dp goal", [subgoal])
  in (qL,pL,rwL)
  end

fun generate_Q fs = let
    val defs = filter (fn (f,_) => substring (f,0,1) = Isafor_Term.def_prefix) fs
    fun mk_xs i = if i = 0 then [] else mk_xs (i - 1) @ [Isafor_Term.mk_var ("x" ^ string_of_int i)]
    fun mk_q (f,n) = Isafor_Term.mk_app true false f (mk_xs n)
    val q = map mk_q defs
          |> HOLogic.mk_list Isafor_Term.termT
    val qs = @{term "List.set :: ((string,string)term list \<Rightarrow> (string,string)term set)"} $ q
  in qs end

fun expand_dp_tac ctxt = Simplifier.simp_tac (put_simpset HOL_basic_ss (Context.proof_of ctxt) addsimps
        @{thms set_list_conv} addsimps (Trs_Data.all_trs_defs ctxt)) 
    THEN' (fn i => fn thm => let
      val (_,p,rw) = extract_dp_problem thm
      val pl = HOLogic.dest_list p
      val rwl = HOLogic.dest_list rw
      val fas = Isafor_Term.funas_trs (pl @ rwl)
      val q = generate_Q fas
            |> Thm.global_cterm_of (Context.theory_of ctxt)
            |> SOME
      val mono_thm = Thm.instantiate' [] [q] @{thm finite_dpp_mono_tf_Q}
    in resolve_tac (Context.proof_of ctxt) [mono_thm] i thm
    end)
    THEN' Simplifier.simp_tac (put_simpset termfun_ss (Context.proof_of ctxt) addsimps @{thms tf_Q_def})

fun expand_trs_tac ctxt = Simplifier.simp_tac (put_simpset HOL_basic_ss (Context.proof_of ctxt) addsimps
        @{thms set_list_conv} addsimps (Trs_Data.all_trs_defs ctxt)) 
    THEN' (fn i => fn thm => let
      val (_,r) = extract_trs_problem thm
      val rl = HOLogic.dest_list r
      val fas = Isafor_Term.funas_trs rl
      val q = generate_Q fas
            |> Thm.global_cterm_of (Context.theory_of ctxt)
            |> SOME
      val mono_thm = Thm.instantiate' [] [q] @{thm SN_trs_mono_tf_Q}
    in resolve_tac (Context.proof_of ctxt) [mono_thm] i thm
    end)
    THEN' Simplifier.simp_tac (put_simpset termfun_ss (Context.proof_of ctxt) addsimps @{thms tf_Q_def})

val setup = 
(*  #> Datatype.interpretation (K register_datatypes) *)(* not working yet *)
  Method.setup @{binding trs_reduce}
    (Args.term >> (fn f => fn ctxt => SIMPLE_METHOD' (Isafor_Tactics.reduce_tac ctxt f)))
    "reduces a termination goal to the termination of a term rewrite system"
  #> Trs_Data.setup
  #> Method.setup @{binding ceta_import_proof}
    (Scan.lift Parse.path >> (fn path => fn ctxt => SIMPLE_METHOD' (ceta_import_proof_tac ctxt (Path.explode path))))
    "reads a termination proof from a file" 
  #> Method.setup @{binding ceta_import_dp_proof}
    (Scan.lift Parse.path >> (fn path => fn ctxt => SIMPLE_METHOD' (ceta_import_dp_proof_tac ctxt (Path.explode path))))
    "reads a finiteness proof from a file" 
  #> Method.setup @{binding dps_reduce}
    (Args.term >> (fn f => fn ctxt => SIMPLE_METHOD' (Isafor_Tactics.reduce_dp_tac ctxt f)))
    "reduces a termination goal to finiteness of a DP problem"
  #> Method.setup @{binding ceta_checker_simp}
    (Scan.succeed (fn ctxt => SIMPLE_METHOD' (ceta_checker_simp_tac ctxt)))
    "evaluates ceta_checker using the simplifier" 
  #> Method.setup @{binding expand_dp_problem}
    (Scan.succeed (fn ctxt => SIMPLE_METHOD' (expand_dp_tac (Context.Proof ctxt))))
    "expands the dp problem and restrict the set Q to those terms occurring in the dp problem" 
  #> Method.setup @{binding expand_trs_problem}
    (Scan.succeed (fn ctxt => SIMPLE_METHOD' (expand_trs_tac (Context.Proof ctxt))))
    "expands the trs problem and restrict the set Q to those terms occurring in the trs" 

fun export_termination_problem path proof_state =
  let
    val {goal=goal_state,...} = Proof.goal proof_state
    val (Const (@{const_name List.set}, _) $ qL,trsL) = extract_trs_problem goal_state
    val problem = Xml.mk_problem qL NONE trsL
  in
    File.write path (Xml.to_string problem)
  end


fun export_dp_problem path proof_state =
  let
    val {goal=goal_state,...} = Proof.goal proof_state
    val (Const (@{const_name List.set}, _) $ qL,pL,rwL) = extract_dp_problem goal_state
    val problem = Xml.mk_problem qL (SOME pL) rwL
  in
    File.write path (Xml.to_string problem)
  end


val parser =
  Scan.optional (Parse.$$$ "(" |-- (Parse.reserved "no_sim" >> K false) --| Parse.$$$ ")") true
  -- Parse.thms1

val _ =
  Outer_Syntax.command @{command_keyword define_trs} "defines term rewrite system corresponding to given simp rules"
    (parser >> (fn (sims,facts) => Toplevel.theory (define_trs sims facts)))

val _ =
  Outer_Syntax.command @{command_keyword define_dps} "defines DP problem corresponding to given simp rules"
    (parser >> (fn (_,facts) => Toplevel.theory (define_dps facts)))


val _ =
  Outer_Syntax.command @{command_keyword export_termination_problem} "write termination problem to file"
    (Parse.path >> (fn path => 
      Toplevel.keep (export_termination_problem (Path.explode path) o Toplevel.proof_of)))

val _ =
  Outer_Syntax.command @{command_keyword export_dp_problem} "write termination problem to file"
    (Parse.path >> (fn path => 
      Toplevel.keep (export_dp_problem (Path.explode path) o Toplevel.proof_of)))

end
